//! \file lzw_adapt1.cpp

// LZW   
//     .
//    ,          
//     
//    

// icq:133575351  email:lawnmower-man@mail.ru

#include<stdio.h>
#include<memory.h>
#include<malloc.h>

#define LZ_END	0		//!<   
//[1..256] -   
#define LZ_START 257	//!<    

#define CAPABLITY 65536			//!<  
#define POLYNOM	0xEDB88320		//!<     (ZIP) 

typedef  unsigned char  BYTE;
typedef  unsigned int   DWORD;	// 32 
typedef  unsigned short CELL;	// 16  (   )


class LZW{
	struct dict_st{
		CELL prefix;	//!< 
		CELL   next;	//!<    
		CELL  child;	//!<   
		CELL  use_next;	//!<     
		CELL  use_prev;	//!< ....
		BYTE    sym;	//!<  
	};	//!<     (0        )
	void RemoveUse(int);			//!<     
	void UpdateList(int code);		//!<      
protected:
	DWORD crc32;	//!<  
	int use_list;	//!<    
	int count;		//!<     
	int free_index;	//!<   ,   
	int bits_count;	//!<     
	dict_st *table;	//!<  
	void Init();					//!<  
	void Free();					//!<   ,     
	void Update(int,BYTE);			//!<  
	void UpdateCRC(BYTE);			//!<   
	int  Find(int,BYTE);			//!<    
public:
	LZW();
	~LZW();
};

/*
void LZW::CheckList()
{
	return;
	int use,it;
	if(!use_list) return;
	it=use_list;
	use=table[it].use;
	do{
//		if(use<table[it].use) _asm int 3
		use=table[it].use;
		if(table[table[it].use_next].use_prev!=it) _asm int 3
		it=table[it].use_next;
	}while(it!=use_list);
}
//*/


class LZW_Encoder:LZW{
	BYTE byte,byte_shift; 
	void StartWrite() {byte=0,byte_shift=1;}
	void StopWrite(FILE* out) {fputc(byte,out);}
	void WriteCode(FILE*,int);		//!<    
public:
	void Pack(FILE*,FILE*);			//!< 
	static void _Encode(FILE*,FILE*);
};


class LZW_Decoder:LZW{
	BYTE byte,byte_shift; 
	char *stack;		//<!    
	int stack_size;		//<!    
	void StartRead() {byte=byte_shift=0;}
	int  ReadCode(FILE*);
	void WriteSym(FILE*,int);		//!<    
public:
	LZW_Decoder();
	~LZW_Decoder();
	void Unpack(FILE*,FILE*);		//!< 
	static void _Decode(FILE*,FILE*);
};



LZW::LZW()
{
	table=new dict_st[CAPABLITY];
}

LZW::~LZW()
{
	delete[]table;
}


void LZW::Init()
{
	free_index=0;
	memset(table,0,sizeof(dict_st)*CAPABLITY);
	for(int n=0; n<256; n++) {
		table[n+1].next=n+2;
		table[n+1].sym=n;
	}
	table[256].sym=255;
	count=LZ_START;
	bits_count=9;
}


void LZW::RemoveUse(int index)
{
	int	next=table[index].use_next;
	int prev=table[index].use_prev;
	table[next].use_prev=prev;
	table[prev].use_next=next;
}


//! \param code   
void LZW::UpdateList(int code)
{
	int next,prev;
	if(table[code].child) _asm int 3
	while(code>=LZ_START) {
		//     (       )
		//table[code].use++;
		if(code<count) 
			RemoveUse(code);
		//       
		next=use_list;
		prev=table[use_list].use_prev;
		table[code].use_next=next;
		table[code].use_prev=prev;
		table[next].use_prev=code;
		table[prev].use_next=code;
		use_list=code;
		//   (  )
		code=table[code].prefix;
	}
}

void LZW::UpdateCRC(BYTE sym)
{
	int n=8;
	crc32^=sym;
	while(n--) crc32=(-int((crc32&1)&POLYNOM))^(crc32>>1);
}


//! \param code  
//! \param sym   
//! \result =true    
void LZW::Update(int code,BYTE sym)
{
	int index;
	if(code==free_index)
		return; 
	if(count<CAPABLITY) index=count; else index=free_index,free_index=0;
	if(index<LZ_START) return;
	table[index].prefix=code;
	table[index].next=table[code].child;
	table[code ].child=index;
	table[index].sym=sym;
	table[index].child=0;
	if(count==LZ_START) use_list=table[index].use_next=table[index].use_prev=index;
	else UpdateList(index);
	if(count<CAPABLITY) count++;
}

void LZW::Free()
{
	//     
	if(free_index) return;
	if(table[table[use_list].use_prev].child) return;
	free_index=table[use_list].use_prev;
	//    
	RemoveUse(free_index);
	//   
	int next=table[free_index].next;
	int index=table[free_index].prefix;
	if(table[index].child==free_index) {table[index].child=next;return;}
	index=table[index].child;
	while(index) {
		if(table[index].next==free_index) {table[index].next=next;return;}
		index=table[index].next;
	}
}


//! \param code  
//! \param sym     
//! \result   (0 -   )
int LZW::Find(int code,BYTE sym)
{
	code=table[code].child;
	while(code) {
		if(table[code].sym==sym) break;
		code=table[code].next;
	}
	return code;
}

void LZW_Encoder::WriteCode(FILE *out,int code)
{
	for(int n=0; n<bits_count; n++) {
		if(!byte_shift) fputc(byte,out),byte_shift=1,byte=0;
		if(code&1) byte|=byte_shift;
		byte_shift<<=1;
		code>>=1;
	}
}

void LZW_Encoder::_Encode(FILE *in,FILE *out)
{
	LZW_Encoder obj;
	obj.Pack(in,out);
}


//! \param in   
//! \param out  
void LZW_Encoder::Pack(FILE *in,FILE *out)
{
	int code,sym,index;
	code=0;
	Init(); StartWrite(); crc32=-1;
	for(;;) {
		sym=fgetc(in); if(sym==EOF) break;
		UpdateCRC(sym);
		if(code) index=Find(code,sym); else 
			index=sym+1;
		if(index) code=index;
		else {
			WriteCode(out,code);
			if(count>=CAPABLITY-1) Free();
			Update(code,sym);
			if(count==(1<<bits_count) && count<CAPABLITY) bits_count++;
			code=sym+1;
		}
	}
	WriteCode(out,code);
	WriteCode(out,LZ_END);
	StopWrite(out);
	fwrite(&crc32,1,sizeof(crc32),out);
}

LZW_Decoder::LZW_Decoder()
{
	stack=0; stack_size=0;
}

LZW_Decoder::~LZW_Decoder()
{
	free(stack);
}



//! \param out   
//! \param code     
void LZW_Decoder::WriteSym(FILE *out,int code)
{
	int size=0;
	for(;;) {
		if(!code) break;
		if(size>=stack_size) {
			stack_size+=((stack_size<<1)+9)&~7;
			stack=(char*)realloc(stack,stack_size);
		}
		stack[size++]=table[code].sym;
		code=table[code].prefix;
	}
	while(size) {
		size--;
		UpdateCRC(stack[size]);
		fputc(stack[size],out);
	}
}


//! \param in   
//! \result  
int LZW_Decoder::ReadCode(FILE *in)
{
	int code=0;
	for(int n=0; n<bits_count; n++) {
		if(!byte_shift) byte=fgetc(in),byte_shift=1;
		if(byte&byte_shift) code|=1<<n;
		byte_shift<<=1;
	}
	return code;
}


void LZW_Decoder::_Decode(FILE *in,FILE *out)
{
	LZW_Decoder obj;
	obj.Unpack(in,out);
}


//! \param in   
//! \param out  
void LZW_Decoder::Unpack(FILE *in,FILE *out)
{
	int n,old_code,sym,new_code;
	old_code=0;
	Init(); StartRead(); crc32=-1;
	for(;;) {
		if(count>=CAPABLITY-1) Free();
		new_code=ReadCode(in);
		if(!new_code) break;
		if(new_code>=count || new_code==free_index) WriteSym(out,old_code),WriteSym(out,sym+1),n=old_code;
		else WriteSym(out,new_code),n=new_code;
		while(n) sym=table[n].sym,n=table[n].prefix;
		if(old_code) {
			Update(old_code,sym);
			if(count>=(1<<bits_count)-1 && count<CAPABLITY-1) bits_count++;
		}
		old_code=new_code;
	}
	DWORD crc;
	fread(&crc,1,sizeof(crc),in);
	if(crc!=crc32) printf("\nCRC32 error!!!\n\a");
}


int main(int count,char **list)
{
	if(count!=4) {
		printf("%s -e infile outfile\n",list[0]);
		printf("%s -d infile outfile\n",list[0]);
		return 1;
	}
	if(list[1][0]!='-' || list[1][1]!='e' && list[1][1]!='d') {
		printf("Syntax error\n\a");
		return 2;
	}
	FILE *input =fopen(list[2],"rb");
	FILE *output=fopen(list[3],"wb");
	if(!input || !output) {
		printf("IO error\n\a");
		return 3;
	}
	switch(list[1][1]) {
	case 'e': LZW_Encoder::_Encode(input,output); break;	// 
	case 'd': LZW_Decoder::_Decode(input,output); break;	// 
	}
	fclose(input);
	fclose(output);
	return 0;
//*/
}


